import { getLanguages, registerLanguage, } from "./languageManager.js";
/**
 * The default color scheme for the syntax highlighter.
 * Values are picked generically based on what we thought looked good.
 */
// TODO: change colors, thes look bad
const defaultColors = {
    identifier: '#7eaacc',
    keyword: '#ff7f0e',
    separator: '#dcdcdc',
    operator: '#2ca02c',
    literals: {
        boolean: '#9467bd',
        number: '#1f77b4',
        string: '#ffbb78',
        null: '#d62728',
    },
    comment: '#808080',
};
/**
 * Gets a language definition for a given language name.
 * @param language The language name to get the language definition for.
 * @returns The language definition.
 */
export function getLanguageDefinition(language) {
    let languages = getLanguages();
    languages = languages.filter((lang) => {
        return lang.name.toUpperCase() === language.toUpperCase();
    });
    if (languages.length > 1) {
        console.warn(`Multiple languages found with the name ${language}, Using the first one found.`);
    }
    if (languages.length === 0) {
        console.warn(`No language found with the name ${language}.`);
        return {
            identifier: [],
            keyword: [],
            separator: [],
            operator: [],
            literals: {
                boolean: [],
                number: [],
                string: [],
                null: [],
            },
            comment: [],
        };
    }
    return languages[0].defenition;
}
/**
 * Highlight code in html syntax
 * @param code The code to be highlighted
 * @param options Options for highlighting
 */
export default function (code, options) {
    // Set defualts if not set.
    options.language = options.language || "javascript";
    options.tailwind = options.tailwind || false;
    options.colors = options.colors || defaultColors;
    // Get the language definition based of the language.
    const languageDefinition = getLanguageDefinition(options.language);
    let C = structuredClone(code);
    let found = [];
    // find all comments
    languageDefinition.comment.forEach((commentMatcher) => {
        for (const comment of C.matchAll(commentMatcher)) {
            // Dont need to ajust index since nothing has been removed from C yet
            found.push({ value: comment, type: "comment", index: comment.index });
        }
        C = C.replaceAll(commentMatcher, "");
    });
    // Find all strings
    languageDefinition.literals.string.forEach((stringMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const string of C.matchAll(stringMatcher)) {
            partialFound.push({ value: string, type: "string", index: string.index });
        }
        for (const string of code.matchAll(stringMatcher)) {
            fullFound.push({ value: string, type: "string", index: string.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "string", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        // We dont need to check overlapping strings because they will both be given the same color anyways.
        C = C.replaceAll(stringMatcher, "");
    });
    // find other literals
    // numeric literals
    languageDefinition.literals.number.forEach((numberMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const number of C.matchAll(numberMatcher)) {
            partialFound.push({ value: number, type: "number", index: number.index });
        }
        for (const number of code.matchAll(numberMatcher)) {
            fullFound.push({ value: number, type: "number", index: number.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "number", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(numberMatcher, "");
    });
    // boolean literals
    languageDefinition.literals.boolean.forEach((booleanMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const boolean of C.matchAll(booleanMatcher)) {
            partialFound.push({ value: boolean, type: "boolean", index: boolean.index });
        }
        for (const boolean of code.matchAll(booleanMatcher)) {
            fullFound.push({ value: boolean, type: "boolean", index: boolean.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "boolean", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(booleanMatcher, "");
    });
    // null literals
    languageDefinition.literals.null.forEach((nullMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const nullLiteral of C.matchAll(nullMatcher)) {
            partialFound.push({ value: nullLiteral, type: "null", index: nullLiteral.index });
        }
        for (const nullLiteral of code.matchAll(nullMatcher)) {
            fullFound.push({ value: nullLiteral, type: "nullLiteral", index: nullLiteral.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "nullLiteral", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(nullMatcher, "");
    });
    // find operators
    languageDefinition.operator.forEach((operatorMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const operator of C.matchAll(operatorMatcher)) {
            partialFound.push({ value: operator, type: "operator", index: operator.index });
        }
        for (const operator of code.matchAll(operatorMatcher)) {
            fullFound.push({ value: operator, type: "operator", index: operator.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "operator", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(operatorMatcher, "");
    });
    // find seperators
    languageDefinition.separator.forEach((separatorMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const separator of C.matchAll(separatorMatcher)) {
            partialFound.push({ value: separator, type: "separator", index: separator.index });
        }
        for (const separator of code.matchAll(separatorMatcher)) {
            fullFound.push({ value: separator, type: "separator", index: separator.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "separator", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(separatorMatcher, "");
    });
    // find keywords
    languageDefinition.keyword.forEach((keywordMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const keyword of C.matchAll(keywordMatcher)) {
            partialFound.push({ value: keyword, type: "keyword", index: keyword.index });
        }
        for (const keyword of code.matchAll(keywordMatcher)) {
            fullFound.push({ value: keyword, type: "keyword", index: keyword.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "keyword", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(keywordMatcher, "");
    });
    // find identifiers
    languageDefinition.identifier.forEach((identifierMatcher) => {
        let fullFound = [];
        let partialFound = [];
        for (const identifier of C.matchAll(identifierMatcher)) {
            partialFound.push({ value: identifier, type: "identifier", index: identifier.index });
        }
        for (const identifier of code.matchAll(identifierMatcher)) {
            fullFound.push({ value: identifier, type: "identifier", index: identifier.index });
        }
        for (let i = 0; i < partialFound.length; i++) {
            if (!fullFound[i]) {
                throw new Error("Something went wrong!");
            }
            if (partialFound[i].value[0] === fullFound[i].value[0]) {
                found.push({ value: partialFound[i].value, type: "identifier", index: fullFound[i].index });
            }
            else {
                fullFound.splice(i, 1);
                i -= 1;
            }
        }
        C = C.replaceAll(identifierMatcher, "");
    });
    // New lines in html are made with <br> tags. So we match newlines seperately.
    let fullFound = [];
    let partialFound = [];
    for (const newline of C.matchAll(/\n\r|\n/g)) {
        partialFound.push({ value: newline, type: "newline", index: newline.index });
    }
    for (const newline of code.matchAll(/\n\r|\n/g)) {
        fullFound.push({ value: newline, type: "newline", index: newline.index });
    }
    for (let i = 0; i < partialFound.length; i++) {
        found.push({ value: partialFound[i].value, type: "newline", index: fullFound[i].index });
    }
    // We want the code to be the same in the source as in the result,
    // so we are also matching any unmatched whitespace.
    fullFound = [];
    partialFound = [];
    for (const whitespace of C.matchAll(/\s+?/g)) {
        partialFound.push({ value: whitespace, type: "whitespace", index: whitespace.index });
    }
    for (const whitespace of code.matchAll(/\s+?/g)) {
        fullFound.push({ value: whitespace, type: "whitespace", index: whitespace.index });
    }
    for (let i = 0; i < partialFound.length; i++) {
        found.push({ value: partialFound[i].value, type: "whitespace", index: fullFound[i].index });
    }
    // For the same reason we will also match any other unmatched characters.
    fullFound = [];
    partialFound = [];
    for (const character of C.matchAll(/\S+/g)) {
        partialFound.push({ value: character, type: "character", index: character.index });
    }
    for (const character of code.matchAll(/\S+/g)) {
        fullFound.push({ value: character, type: "character", index: character.index });
    }
    for (let i = 0; i < partialFound.length; i++) {
        found.push({ value: partialFound[i].value, type: "character", index: fullFound[i].index });
    }
    // At this point, everything should be matched and removed.
    // If not the code has a syntax error.
    if (C.trim().length > 0) {
        console.warn(`Syntax error in code: ${code}\nNot everything was matched!`);
    }
    // Sort all found values by where in the code they should be.
    found = found.sort((a, b) => {
        return (a.index || 0) - (b.index || 0);
    });
    let highlighted = "";
    // Color the text
    for (const part of found) {
        switch (part.type) {
            case "newline":
                highlighted += "<br>";
                break;
            case "whitespace":
            case "character":
                highlighted += part.value[0];
                break;
            case "identifier":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.identifier}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.identifier}">${part.value[0]}</span>`;
                break;
            case "keyword":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.keyword}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.keyword}">${part.value[0]}</span>`;
                break;
            case "separator":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.separator}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.separator}">${part.value[0]}</span>`;
                break;
            case "operator":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.operator}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.operator}">${part.value[0]}</span>`;
                break;
            case "boolean":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.literals.boolean}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.literals.boolean}">${part.value[0]}</span>`;
                break;
            case "number":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.literals.number}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.literals.number}">${part.value[0]}</span>`;
                break;
            case "string":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.literals.string}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.literals.string}">${part.value[0]}</span>`;
                break;
            case "null":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.literals.null}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.literals.null}">${part.value[0]}</span>`;
                break;
            case "comment":
                !options.tailwind
                    ? highlighted += `<span style="color: ${options.colors.comment}">${part.value[0]}</span>`
                    : highlighted += `<span class="text-${options.colors.comment}">${part.value[0]}</span>`;
                break;
            default:
                throw new Error("Something went wrong!");
        }
    }
    console.log(highlighted);
    return highlighted;
}
export { registerLanguage };
/**
 * Importing default languages.
 * Any non default language can be imported in the file that uses it.
*/
import "./JavaScript.js";
